# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

import math
from PyQt5 import QtCore


def tangentCalc(circleParams, point):
    """Given a circle defined by circleParams (a dict containing the radius
    and center coordinates of the circle along with the errors on these
    parameters) and point, one of the TrackMarker objects to which
    the circle was fitted, return three QLineF objects, one being the tangent
    to the circle at point and the other two being the errors on the
    tangent line in each direction. The real tangent lies somewhere between
    the two error lines.
    """

    # compute the angular coordinate of the point with respect to the center
    # of the circle
    pointAngle = point.getAngle((circleParams['centerX'],
                                 circleParams['centerY']))*math.pi/180

    # determine the coordinates of the point that has the same angular
    # coordinate as point but actually lies on the circle
    tangentPointX = circleParams['radius']*math.cos(pointAngle) + \
                    circleParams['centerX']
    # note: y values increase going down
    tangentPointY = -circleParams['radius']*math.sin(pointAngle) + \
                    circleParams['centerY']

    # determine the coordinates of the tangent point relative to the center of
    # the circle -- the line joining the centre to the tangent point is
    # perpendicular to the tangent line
    relativeX = tangentPointX - circleParams['centerX']
    relativeY = tangentPointY - circleParams['centerY']

    # construct the tangent line
    tangentLine = QtCore.QLineF(tangentPointX + relativeY,
                                tangentPointY - relativeX,
                                tangentPointX - relativeY,
                                tangentPointY + relativeX)

    # construct the two error tangent lines using the error on the circle
    # center. These two lines also pass through the tangent point and the real
    # tangent line lies between these two lines within error. Note that the
    # error on the circle radius is not considered when constructing the error
    # tangents because it would only translate the tangent radially without
    # changing the slope
    tangentLineErrA = QtCore.QLineF(tangentPointX + \
                                      (relativeY - circleParams['centerYErr']),
                                    tangentPointY - \
                                      (relativeX + circleParams['centerXErr']),
                                    tangentPointX - \
                                      (relativeY - circleParams['centerYErr']),
                                    tangentPointY + \
                                      (relativeX + circleParams['centerXErr']))
    tangentLineErrB = QtCore.QLineF(tangentPointX + \
                                      (relativeY + circleParams['centerYErr']),
                                    tangentPointY - \
                                      (relativeX - circleParams['centerXErr']),
                                    tangentPointX - \
                                      (relativeY + circleParams['centerYErr']),
                                    tangentPointY + \
                                      (relativeX - circleParams['centerXErr']))

    return tangentLine, tangentLineErrA, tangentLineErrB

def openingAngle(tangent, tangentErrA, tangentErrB, refLine):
    """Return the angle (in degrees) between tangent, a QLineF, and refLine,
    a ReferenceLine object. Return also the error on the angle using the
    two tangent error lines tangentErrA and tangentErrB, both QLineF objects.
    """

    # compute the angle between the reference line and the tangent line
    angle = refLine.line.line().angleTo(tangent)

    # compute the differences between the angles between the reference line
    # and the error tangent lines and the angle between the reference line
    # and the tangent estimate. These give the error on the angle in both
    # directions.
    angleErrA = abs(refLine.line.line().angleTo(tangentErrA) - angle)
    angleErrB = abs(refLine.line.line().angleTo(tangentErrB) - angle)

    # special case: reference line is in between the two error tangents.
    # one of the two angle errors will be 360 - theta in this case
    if angleErrA > 180:
        angleErrA = abs(angleErrA - 360)
    elif angleErrB > 180:
        angleErrB = abs(angleErrB - 360)

    # take the error on the angle to be the average of the errors in each
    # direction
    errAngle = (angleErrA + angleErrB) / 2

    return angle, errAngle



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
from PyQt5 import QtWidgets, QtGui
from traxis.gui import maingui


class TraxisApplicationWindow(QtWidgets.QMainWindow):

    """Subclass of QMainWindow, which sets the window's title and icon, and
    instantiates the MainWidget, setting it as the window's central widget.
    """

    def __init__(self):

        super().__init__()

        # set window title
        self.setWindowTitle("Traxis")

        # set the window's icon (it is kept in the root traxis directory)
        # note: this icon won't get displayed on a Mac
        icon = QtGui.QIcon()
        basePath = sys.path[0]
        icon.addPixmap(QtGui.QPixmap(os.path.join(basePath, "traxis.png")))
        self.setWindowIcon(icon)

        # initialize MainWidget and set it as the window's central widget
        self.setCentralWidget(maingui.MainWidget())



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtGui

# CALIBRATION CONSTANTS
# ---------------------

# Resolution
CMPERPX = 0.01188 # the number of cm in the actual bubble chamber corresponding
                  # to each pixel in a bubble chamber image scan
ERRCMPERPX = 0.00090 # the error on the px to cm conversion

# bubble chamber magnetic field
MAGNETICFIELD = 15.5 # in kG

# speed of light
C = 0.3 # in giga metres per second


# INTERFACE CONFIGURATION
# -----------------------

# marker colours
DEFAULTMARKERCOLOR = QtGui.QColor(176, 30, 125)
STARTMARKERCOLOR = QtGui.QColor(0, 186, 186)
ENDMARKERCOLOR = QtGui.QColor(34, 197, 25)
HIGHLIGHTMARKERCOLOR = QtGui.QColor(235, 233, 0)

# momentum arc colours
ARCCOLOR = QtGui.QColor(33, 95, 147)

# reference line colours
REFLINECOLOR = QtGui.QColor(243, 42, 31)

# tangent line colours
TANGENTLINECOLOR = QtGui.QColor(0, 0, 0)

# zoom factors
ZOOMINFACTOR = 1.25 # this should be greater than 1
ZOOMOUTFACTOR = 1/ZOOMINFACTOR

# default GUI state variables
DEFAULTPOINTSIZE = 10
DEFAULTLINEWIDTH = 2.5



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy import optimize

import traxis.constants as constants

def _createXYArrays(markerList):
    """Given markerList, a MarkerList object, return two numpy arrays, one
    containing the x-coordinates of the markers and the second the
    corrsponding y-coordinates.
    """

    # initialize two numpy arrays whose lengths are equal to the number of
    # markers in markerList
    xArray = np.zeros(markerList.count())
    yArray = np.zeros(markerList.count())

    # loop over each marker, setting the x-coordinate of each marker as an
    # element of xArray and each y-coordinate as an element of yArray
    for row in range(markerList.count()):
        marker = markerList.item(row)
        # using 64 bit floats seems to be necessary to ensure consistent
        # results across all systems. Some systems get the same results
        # with or without np.float64() but others don't. When using
        # float64, results match for all systems tested.
        xArray[row] = np.float64(marker.ellipse.rect().center().x())
        yArray[row] = np.float64(marker.ellipse.rect().center().y())

    return xArray, yArray

def _distanceResiduals(referencePoint, xArray, yArray):
    """Given referencePoint, a tuple whose first element is the x-coordinate
    and whose second element is the y-coordinate, xArray, an array of
    x-coordinates, and yArray, an array of the corresponding y-coordinates of
    a set of points (xArray and yArray are expected to have the same length),
    return an array containing for each point the difference between the
    distance from that point to the reference point and the mean of the
    distances.
    """

    # compute the distance from each point to the reference point
    distances = np.sqrt((xArray - referencePoint[0])**2 + \
                        (yArray - referencePoint[1])**2)
    # compute the difference between the distance from each point to the
    # reference point and the mean of the distances
    distanceResiduals = distances - distances.mean()

    return distanceResiduals

def fitCircle(markerList):
    """Given markerList, a MarkerList object, fit a circle to the markers
    using the least squares method and return the coordinates of the centre
    and the radius of the fitted circle along with their errors.
    """

    # convert the list of track markers to arrays of the x-coordinates and
    # y-coordinates of the markers
    xArray, yArray = _createXYArrays(markerList)

    # use the mean of the x-coordinates of the markers as an initial guess for
    # x-coordinate of the circle center. Similarly for y-coordinates.
    centerEstimate = (xArray.mean(), yArray.mean())

    # initialize the dictionary containing the parameters of the fitted circle
    # to be returned by this function
    fitParams = {}

    # calculate the optimal centre coordinates for the fitted circle, such
    # that the squares of the residuals of the distances from each point to
    # the centre are minimized. Store the optimal coordinates in the variable
    # centerLsq. Pass full_output=True to the leastsq function so that the
    # covariance matrix of the optimized coordinates is returned. This also
    # makes the function return infodict, mesg and ier which are not used.
    # Decrease the desired error in the results by passing explicit values
    # for ftol and xtol (the default value is 1.49012e-8)
    centerLsq, covMatrix, _, _, _ = optimize.leastsq(
                                                    _distanceResiduals,
                                                    centerEstimate, 
                                                    args=(xArray, yArray), 
                                                    full_output=True, 
                                                    ftol=1e-15, xtol=1e-15)

    # note from the documentation for scipy.optimize.leastsq regarding the
    # covariance matrix it returns: "This matrix must be multiplied by the
    # residual variance to get the covariance of the parameter estimates"
    # The 'residual variance' is just chi-squared/degrees of freedom
    dof = len(yArray) - len(centerEstimate)
    chi2Dof = (_distanceResiduals(centerLsq, xArray, yArray)**2).sum()/dof
    # multiply covMatrix by the 'residual variance' to get the covariance
    # of the parameter estimates
    parameterCov = covMatrix * chi2Dof

    # store the optimal centre coordinates in the fitParams dict
    fitParams['centerX'], fitParams['centerY'] = centerLsq
    # store the errors on the optimal centre coordinates in the fitParams dict.
    # the errors are the square roots of the diagonal elements of the parameter
    # covariance
    fitParams['centerXErr'], fitParams['centerYErr'] = np.sqrt(np.diag(parameterCov))

    # store the radius of the fitted circle in the fitParams dict. Based on
    # how the lsq fit was performed, this is the mean of the distances from
    # each point to the optimal circle centre
    fitParams['radius'] = np.sqrt((xArray - fitParams['centerX'])**2 + \
                                  (yArray - fitParams['centerY'])**2).mean()

    # store the error on the radius in the fitParams dict. Given radius R,
    # centre coordinates (h, k) with error dh, dk and a point (x, y) assumed
    # to be known to perfect accuracy, the error on the radius is
    # dR^2 = ((x-h)/R * dh)^2 + ((y-k)/R * dk)^2
    # Compute this dR for each point individually and then take the mean
    fitParams['radiusErr'] = np.sqrt(
        ((fitParams['centerX']-xArray).mean()/ \
         fitParams['radius']*fitParams['centerXErr'])**2 + \
        ((fitParams['centerY']-yArray).mean()/ \
         fitParams['radius']*fitParams['centerYErr'])**2)

    return fitParams

def calc_momentum(radius, uncertainty):
    """ Calculate the track momentum and errors from a fitted circle.

    Given a track radius, R, in cm, a magnetic field, B,
    in kG and the speed of light, c, in Giga metres per second, the
    track momentum in MeV/c can be computed as p = c*B*R

    Parameters
    ----------
    radius      : The track radius in pixels
    uncertainty : the statistical uncertainty of the track radius

    Returns
    -------
    momemtum    : the calculated momentum
    stat_unc    : the statistical uncertainty from the fit
    cal_unc     : the uncertainty due to calibration
    """
    factor = constants.C * constants.MAGNETICFIELD * constants.CMPERPX
    momentum = factor * radius
    stat_unc = factor * uncertainty
    cal_unc = momentum / constants.CMPERPX * constants.ERRCMPERPX
    return momentum, stat_unc, cal_unc




# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import ctypes
from PyQt5 import QtWidgets
from traxis.gui import mainwindow


def main():
    # if running on Windows, create an application user model id for this app
    # so that a custom taskbar icon can be used
    if hasattr(ctypes, 'windll'):
        appId = 'traxis'
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appId)

    # store the path of this file at the start of sys.path so that all files under
    # the root traxis directory can be easily accessed
    basePath = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, basePath)

    # create a QApplication object for managing the GUI control flow and settings
    app = QtWidgets.QApplication(sys.argv)

    # create an instance of TraxisApplicationWindow and display it on the screen
    window = mainwindow.TraxisApplicationWindow()
    window.show()

    # begin the app's event handling loop; ensure a clean exit
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtWidgets, QtGui, QtCore
from traxis import constants


class ReferenceLine(object):
    """Angle reference line class. This class is a container for two
    QGraphicsEllipseItem objects and one QGraphicsLineItem object
    which together form an angle reference. This class implements
    methods for manipulating the graphics items it contains such
    that they function as a unit. A ReferenceLine object is to be
    added to a GUI as one would add a QWidget.
    """

    def __init__(self):
        """Initialize this object's attributes, setting them to None."""

        # the initialPoint attribute is a QGraphicsEllipseItem or None, if no
        # initialPoint has been created
        self.initialPoint = None
        # the finalPoint attribute is a QGraphicsEllipseItem or None, if no
        # finalPoint has been created
        self.finalPoint = None
        # the line attribute is a QGraphicsLineItem or None, if no
        # line has been created
        self.line = None

    def setInitialPoint(self, x, y, size, width, scene):
        """First, reset this reference line object since setting a new initial
        point means that a new reference line is being created. Then create a 
        new ellipse item at position (x, y) with size size and pen width width.
        Set the ellipse as this reference line's initialPoint attribute and add
        the newly created ellipse to scene, a QGraphicsScene.
        """

        # reset this reference line object
        self.reset()

        # set a minimum rect size
        if size < 1:
            size = 1

        # set a minimum pen width
        if width < 1:
            width = 1
        # Force integer for Qt5 compatibility
        width = int(width)

        # create a rect for the initial point with the given
        # coordinates and size (set the width and height of the rect to size)
        initialRect = QtCore.QRectF(x, y, size, size)
        # when instantiating a rect, the given coordinates are used for the
        # top-left corner so we must explicitly move the rect's center to the
        # desired coordinates
        initialRect.moveCenter(QtCore.QPointF(x, y))
        # create a QGraphicsEllipseItem using the rect created above. Set it
        # as the reference line's initialPoint attribute.
        self.initialPoint = QtWidgets.QGraphicsEllipseItem(initialRect)

        # create a pen for the initial point using the reference line colour
        initialPen = QtGui.QPen(constants.REFLINECOLOR)
        # set the width of the pen to width
        initialPen.setWidth(width)
        # set the newly created pen as the inital point's pen
        self.initialPoint.setPen(initialPen)

        # add the initial point to the graphics scene
        scene.addItem(self.initialPoint)

    def setFinalPoint(self, x, y, size, width, scene):
        """Create a new ellipse item at position (x, y) with size size and pen
        width width. Set the ellipse as this reference line's finalPoint
        attribute and add the newly created ellipse to scene, a QGraphicsScene.
        """

        # set a minimum rect size
        if size < 1:
            size = 1

        # set a minimum pen width
        if width < 1:
            width = 1
        width = int(width)

        # create a rect for the final point with the given
        # coordinates and size (set the width and height of the rect to size)
        finalRect = QtCore.QRectF(x, y, size, size)
        # when instantiating a rect, the given coordinates are used for the
        # top-left corner so we must explicitly move the rect's center to the
        # desired coordinates
        finalRect.moveCenter(QtCore.QPointF(x, y))
        # create a QGraphicsEllipseItem using the rect created above. Set it
        # as the reference line's finalPoint attribute.
        self.finalPoint = QtWidgets.QGraphicsEllipseItem(finalRect)

        # create a pen for the initial point using the reference line colour
        finalPen = QtGui.QPen(constants.REFLINECOLOR)
        # set the width of the pen to width
        finalPen.setWidth(width)
        # set the newly created pen as the final point's pen
        self.finalPoint.setPen(finalPen)

        # add the final point to the graphics scene
        scene.addItem(self.finalPoint)

        # if the coordinates of the final point are the same as the initial
        # point, reset the reference line altogether
        if self.finalPoint.rect().center() == self.initialPoint.rect().center():
            self.reset()

    def drawLine(self, endX, endY, width, scene):
        """First, remove the existing line item, if any, from the graphics
        scene. Then create a new line item whose start point coordinates are
        the coordinates of initialPoint, whose end point coordinates are
        (endX, endY), and with pen width width. Add the newly created line
        to scene, a QGraphicsScene. 
        """

        # if this reference line already has a line attribute, remove it from
        # the graphics scene
        if self.line:
            scene.removeItem(self.line)

        # set a minimum pen width
        if width < 1:
            width = 1
        width = int(width)

        # create a QGraphicsLineItem with the appropriate start and end
        # coordinates
        startX = self.initialPoint.rect().center().x()
        startY = self.initialPoint.rect().center().y()
        self.line = QtWidgets.QGraphicsLineItem(startX, startY, endX, endY)

        # create a pen for the line using the reference line colour
        linePen = QtGui.QPen(constants.REFLINECOLOR)
        # set the width of the pen to width
        linePen.setWidth(width)
        # set the newly created pen as the line's pen
        self.line.setPen(linePen)

        # add the line to the graphics scene
        scene.addItem(self.line)

    def rescale(self, size, width):
        """Set the size of the initial and final point to size and set the pen
        width of the initial and final point and line to width.
        """

        # set a minimum rect size
        if size < 1:
            size = 1

        # set a minimum pen width
        if width < 1:
            width = 1
        width = int(width)

        # if the initialPoint attribute is not None
        if self.initialPoint:
            # create a new rect, starting from the existing rect of the initial
            # point
            newInitialRect = self.initialPoint.rect()
            # set the width and height of the new rect
            newInitialRect.setWidth(size)
            newInitialRect.setHeight(size)
            # move the rect's center to match that of the original rect's (the
            # center gets shifted when resizing)
            newInitialRect.moveCenter(self.initialPoint.rect().center())
            # set the resized rect as the initial point's rect
            self.initialPoint.setRect(newInitialRect)

            # create a new pen, starting from the existing pen of the initial
            # point
            newInitialPen = self.initialPoint.pen()
            # set the width of the new pen
            newInitialPen.setWidth(width)
            # set the resized pen as the initial point's pen
            self.initialPoint.setPen(newInitialPen)

        # if the finalPoint attribute is not None
        if self.finalPoint:
            # create a new rect, starting from the existing rect of the final
            # point
            newFinalRect = self.finalPoint.rect()
            # set the width and height of the new rect
            newFinalRect.setWidth(size)
            newFinalRect.setHeight(size)
            # move the rect's center to match that of the original rect's (the
            # center gets shifted when resizing)
            newFinalRect.moveCenter(self.finalPoint.rect().center())
            # set the resized rect as the final point's rect
            self.finalPoint.setRect(newFinalRect)

            # create a new pen, starting from the existing pen of the final
            # point
            newFinalPen = self.finalPoint.pen()
            # set the width of the new pen
            newFinalPen.setWidth(width)
            # set the resized pen as the final point's pen
            self.finalPoint.setPen(newFinalPen)

        # if the line attribute is not None
        if self.line:
            # create a new pen, starting from the existing pen of the line
            newLinePen = self.line.pen()
            # set the width of the new pen
            newLinePen.setWidth(width)
            # set the resized pen as the line's pen
            self.line.setPen(newLinePen)

    def isBeingDrawn(self):
        """Return True if the ReferenceLine object is in the process of being
        drawn, False otherwise.
        """

        # line is in process of being drawn if the initial point exists but the
        # final point does not
        return self.initialPoint and not self.finalPoint

    def reset(self):
        """Remove the reference line's attributes from their graphics scene and
        reset them to None.
        """

        # if the initial point, final point and/or line exist, remove them from
        # their graphics scene
        if self.initialPoint:
            self.initialPoint.scene().removeItem(self.initialPoint)
        if self.finalPoint:
            self.finalPoint.scene().removeItem(self.finalPoint)
        if self.line:
            self.line.scene().removeItem(self.line)

        # reset the reference line's attributes to None
        self.initialPoint = None
        self.finalPoint = None
        self.line = None



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtWidgets, QtGui, QtCore
from traxis import constants


class MarkerList(QtWidgets.QListWidget):
    """Track Marker list class.

    This class is intended to contain TrackMarker objects, and implements a
    number of methods to work with the extra features that TrackMarker adds to
    QListWidgetItem.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

    def addMarker(self, x, y, size, width, scene):
        """Create a new TrackMarker and add it to the list.

        The marker is created with this list as its parent, and is drawn to the
        designated QGraphicsScene.

        Parameters
        ----------
        x, y  : (floats) marker coordinates
        size  : size of ellipse to draw
        width : pen width
        scene : the QGraphicsScene to draw the marker on.

        Returns
        -------
        the TrackMarker object.
        """
        # use the last item in the list to determine the new marker's id
        lastItem = self.item(self.count()-1)
        if lastItem:
            newMarkerId = lastItem.id + 1
        else:
            newMarkerId = 1

        newMarker = TrackMarker(newMarkerId, x, y, size, width, self)
        self.setCurrentItem(newMarker)
        scene.addItem(newMarker.ellipse)

        return newMarker

    def deleteMarker(self, marker):
        """Remove the TrackMarker object marker from this marker list."""
        # remove the marker's ellipse from its graphics scene
        marker.ellipse.scene().removeItem(marker.ellipse)

        markerRow = self.row(marker)
        self.takeItem(markerRow)

    def empty(self):
        """Remove all TrackMarker objects from this marker list."""
        # First remove the markers from the scene
        for row in range(self.count()):
            marker = self.item(row)
            marker.ellipse.scene().removeItem(marker.ellipse)
        # now it's safe to clear the list
        self.clear()

    def rescale(self, size, width):
        """Set the size of each marker's circle.

        Parameters
        size  : the circle's diameter
        width : the pen width
        """
        for row in range(self.count()):
            marker = self.item(row)
            marker.rescale(size, width)

    def setStartPoint(self, marker):
        """Designate marker as the start point."""
        # Clear the old startpoint's designation
        oldStartPoint = self.getStartPoint()
        if oldStartPoint:
            oldStartPoint.setDesignation()
            oldStartPoint.recolor()

        # designate marker as the new start point
        marker.setDesignation('start')

    def setEndPoint(self, marker):
        """Designate marker as the end point for this list of markers."""
        # Clear the old startpoint's designation
        oldEndPoint = self.getEndPoint()
        if oldEndPoint:
            oldEndPoint.setDesignation()
            oldEndPoint.recolor()

        # designate marker as the new start point
        marker.setDesignation('end')

    def getStartPoint(self):
        """Return the TrackMarker object designated as the start point

        If no TrackMarker is designated as the start, return None.
        """
        for row in range(self.count()):
            marker = self.item(row)
            if marker.designation == 'start':
                return marker
        return None

    def getEndPoint(self):
        """Return the TrackMarker object designated as the end point

        If no TrackMarker is designated as the end, return None.
        """
        for row in range(self.count()):
            marker = self.item(row)
            if marker.designation == 'end':
                return marker
        return None

    def highlightCurrent(self):
        """Change all markers to their correct colour."""
        for row in range(self.count()):
            marker = self.item(row)
            marker.recolor()

    def selectNext(self):
        """Set the currently selected marker to the next marker in the list."""
        # if there are no markers selected (currentRow() == -1) or if the
        # currently selected marker is the last one in the list, return
        if self.currentRow() == -1 or self.currentRow() == self.count() - 1:
            return
        else:
            self.setCurrentRow(self.currentRow() + 1)

    def selectPrevious(self):
        """Set the currently selected marker to the previous marker in the list."""
        # if there are no markers selected (currentRow() == -1) or if the
        # currently selected marker is the first one in the list, return
        if self.currentRow() == -1 or self.currentRow() == 0:
            return
        else:
            self.setCurrentRow(self.currentRow() - 1)

class TrackMarker(QtWidgets.QListWidgetItem):
    """Track marker class.

    This class subclasses QListWidgetItem, adding attributes for markers:
       - id (unique),
       - displayed name
       - designation (start, end, selected, or none),
       - coordinates
       - QGraphicsEllipseItem
    """

    def __init__(self, markerId, x, y, size, width, parent=None):
        """ Create a new marker.

        It is up to the caller to ensure the markerId is unique. This is
        primarily used to create a displayed name.

        Parameters:
        markerId    : identifier for the marker,
        x, y        : (float) x and y coordinates,
        size        : (float) the size of the marker,
        width       : (float) the width of the pen used to draw the marker,
        parent      : (optional) MarkerList to which the marker will be added.
        """
        self.id = markerId
        self._x = x
        self._y = y

        self.designation = None
        markerName = "Point {}".format(self.id)

        super().__init__(markerName, parent)

        # set a minimum rect size
        if size < 2:
            size = 2

        # set a minimum pen width
        if width < 1:
            width = 1
        width = int(width)

        # create the ellipse to be drawn as the marker
        ellipsePen = QtGui.QPen(constants.DEFAULTMARKERCOLOR)
        ellipsePen.setWidth(width)
        ellipseRect = QtCore.QRectF(x, y, size, size)
        ellipseRect.moveCenter(QtCore.QPointF(x, y))

        self.ellipse = QtWidgets.QGraphicsEllipseItem(ellipseRect)
        self.ellipse.setPen(ellipsePen)

    @property
    def x(self):
        return self._x
    @x.setter
    def x(self, value):
        self._x = value

    @property
    def y(self):
        return self._y
    @y.setter
    def y(self, value):
        self._y = value

    def setDesignation(self, designation=None):
        """Indicate the marker as a Start Point, an End Point or neither, as
        specified by designation (a string).
        """
        if designation not in [None, 'start', 'end']:
            return

        self.designation = designation

        # indicate the designation in the displayed name
        if designation == 'start':
            self.setText("s - Point {}".format(self.id))
        elif designation == 'end':
            self.setText("e - Point {}".format(self.id))
        else:
            self.setText("Point {}".format(self.id))

    def recolor(self):
        """Update the colour of the marker based on its designation or whether
        it is currently selected.
        """
        newPen = self.ellipse.pen()

        # Set the pen's colour based on the designation
        if self.isSelected():
            newPen.setColor(constants.HIGHLIGHTMARKERCOLOR)
        elif self.designation == 'start':
            newPen.setColor(constants.STARTMARKERCOLOR)
        elif self.designation == 'end':
            newPen.setColor(constants.ENDMARKERCOLOR)
        else:
            newPen.setColor(constants.DEFAULTMARKERCOLOR)

        self.ellipse.setPen(newPen)

    def move(self, dx, dy):
        """Move the marker from its current position (x, y) to (x+dx, y+dy).
        dx and dy are floats.
        """
        # create a new rect, starting from the existing rect of the marker's
        # ellipse
        newRect = self.ellipse.rect()
        # translate the rect by (dx, dy)
        newRect.translate(dx, dy)
        # set the translated rect as the marker's ellipse's rect
        self.ellipse.setRect(newRect)

    def getAngle(self, origin, referenceMarker=None):
        """Return the marker's angular coordinate.

        Parameters:
        origin          : a tuple of x, and y values for the origin.
        referenceMarker : another marker to use as an angle reference with the origin

        Returns:
        angle : The marker's angle in polar coordinates using the origin, or
                relative to the line connecting the origin to tyhe reference marker.
        """
        markerVector = QtCore.QLineF(origin[0], origin[1], self.x, self.y)

        # define the the vector to measure the angle from
        if referenceMarker:
            referenceX = referenceMarker.x
            referenceY = referenceMarker.y
        else:
            referenceX = origin[0] + 1
            referenceY = origin[1]
        referenceVector = QtCore.QLineF(origin[0], origin[1], referenceX, referenceY)

        # compute the angular coordinate of the marker
        angle = referenceVector.angleTo(markerVector)

        return angle

    def rescale(self, size, width):
        """Set the radius and penwidth of the marker's circle."""
        if size < 2:
            size = 2
        if width < 1:
            width = 1
        width = int(width)

        # Make the new circle by updating ellipse's rectangle
        newRect = self.ellipse.rect()
        newRect.setWidth(size)
        newRect.setHeight(size)
        # move the rect's center to match that of the original rect's (the
        # center gets shifted when resizing)
        newRect.moveCenter(self.ellipse.rect().center())
        self.ellipse.setRect(newRect)

        # Set the circle's penwidth
        newPen = self.ellipse.pen()
        newPen.setWidth(width)
        self.ellipse.setPen(newPen)



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
# Minor modification October 2017 by Saila Shama
#	Fixed a bug in maingui.py where load session crashes the session
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

import json
import math
from PyQt5 import QtWidgets, QtGui, QtCore
from traxis import constants
from traxis.gui import skeleton
from traxis.calc import anglecalc, circlefit, optdensity
from traxis.graphics import tangent


class MainWidget(skeleton.GuiSkeleton):
    """Class that extends the base skeleton widget class, implementing
    logic that connects button clicks and other widget events to their
    handlers. Defines the GUI's event handler methods.
    """

    def __init__(self):
        """Initialize the GUI skeleton, set default state variables and connect
        buttons and events to methods.
        """

        # initialize the skeleton of the GUI
        super().__init__()

        # set default GUI state variables
        self.msgNumber = 0
        self.zoomFactor = 1
        self.pointSize = constants.DEFAULTPOINTSIZE
        self.lineWidth = constants.DEFAULTLINEWIDTH
        self.imageFileName = None

        # connect buttons
        self.openImageButton.clicked.connect(self.openImage)
        self.saveSessionButton.clicked.connect(self.saveSession)
        self.loadSessionButton.clicked.connect(self.loadSession)
        self.screenshotButton.clicked.connect(self.saveScreenshot)
        self.zoomInButton.clicked.connect(self.zoomIn)
        self.zoomOutButton.clicked.connect(self.zoomOut)
        self.resetButton.clicked.connect(self.reset)
        self.calcMomentumButton.clicked.connect(self.calcTrackMomentum)
        self.calcDensityButton.clicked.connect(self.calcOptDensity)
        self.calcAngleButton.clicked.connect(self.calcAngle)
        self.placeMarkerButton.clicked.connect(self.placeMarkerButtonFunc)
        self.drawRefButton.clicked.connect(self.drawRefButtonFunc)

        # connect the scene pixmap's mouse events
        self.scenePixmap.mousePressEvent = self.pixmapMousePress
        self.scenePixmap.mouseReleaseEvent = self.pixmapMouseRelease
        self.scenePixmap.mouseMoveEvent = self.pixmapMouseMove

        # connect other events
        self.dlLineEdit.textEdited.connect(self.dLEdited)
        self.markerList.itemSelectionChanged.connect(self.highlightPoint)

    ##############################
    # Keypress Event Handler
    ##############################
    def keyPressEvent(self, event):
        """Handle the key presses that are not hooked up to buttons."""

        # get the currently selected track marker
        currentPoint = self.markerList.currentItem()
        # initialize the point movement variables
        dx, dy = 0, 0

        # check if the Shift key was held
        if event.modifiers() & QtCore.Qt.ShiftModifier:
            isShift = True
        else:
            isShift = False

        # WASD to set the direction of the movement of the selected point
        if event.key() == QtCore.Qt.Key_W:
            dy = -1 # up (y increases going down)

        elif event.key() == QtCore.Qt.Key_S:
            dy = 1 # down (y increases going down)

        elif event.key() == QtCore.Qt.Key_D:
            dx = 1 # right

        elif event.key() == QtCore.Qt.Key_A:
            dx = -1 # left

        # if one of WASD was pressed, move the currently selected point
        if dx or dy:
            # if shift was held, do a course movement (half the point size, if
            # the point size is more than 2 px), otherwise move the point by
            # 1 px
            if isShift and self.pointSize >= 2:
                dx *= self.pointSize / 2
                dy *= self.pointSize / 2
            if currentPoint:
                currentPoint.move(dx, dy)
        
        # F/V to select the next or previous marker in the marker list
        elif event.key() == QtCore.Qt.Key_V:
            self.markerList.selectNext()

        elif event.key() == QtCore.Qt.Key_F:
            self.markerList.selectPrevious()

        # G/H to set the currently selected point as the start or end point
        elif event.key() == QtCore.Qt.Key_G:
            if currentPoint:
                self.markerList.setStartPoint(currentPoint)

        elif event.key() == QtCore.Qt.Key_H:
            if currentPoint:
                self.markerList.setEndPoint(currentPoint)

        # Delete key to delete the currently selected point
        elif event.key() == QtCore.Qt.Key_Delete:
            if currentPoint:
                self.markerList.deleteMarker(currentPoint)

    #############################
    # Pixmap Mouse Event Handlers
    #############################
    def pixmapMousePress(self, event):
        """Respond differently to mouse presses on the image depending on the
        currently selected mode. If in track marker placment mode, place a
        marker, if in angle reference drawing mode, reset the reference line
        and set the initial point for a new line, if in neither mode,
        initialize an image pan. event is a QGraphicsSceneMouseEvent object
        containing the coordinates of the mouse press.
        """

        # if the track marker placement mode is selected, add a new marker at
        # the location of the mouse press
        if self.placeMarkerButton.isChecked():
            self.markerList.addMarker(
                event.pos().x(), event.pos().y(), 
                self.pointSize, self.lineWidth, self.scene)

        # if angle reference drawing mode is selected, set the initial point
        # of the reference line at the location of the mouse press
        elif self.drawRefButton.isChecked():
            self.angleRefLine.setInitialPoint(
                event.pos().x(), event.pos().y(),
                self.pointSize, self.lineWidth, self.scene)

        # if neither mode is selected, set the initial reference position for
        # an image pan (store the position of this mouse press)
        # note: the event position is in image/scene coordinates but the pan
        # is to be done using graphics view coordinates so apply the
        # coordinate mapping
        else:
            self.sceneView.lastMousePos = self.sceneView.mapFromScene(
                                              event.pos())

    def pixmapMouseRelease(self, event):
        """Set the final point of the angle reference line if it is in the
        process of being drawn. event is a QGraphicsSceneMouseEvent object
        containing the coordinates of the mouse release.
        """

        # if the reference line is being drawn, set the final point of the
        # reference line at the location of the mouse release 
        if self.angleRefLine.isBeingDrawn():
            self.angleRefLine.setFinalPoint(
                event.pos().x(), event.pos().y(),
                self.pointSize, self.lineWidth, self.scene)

    def pixmapMouseMove(self, event):
        """Redraw the line of the angle reference if it is in the process of
        being drawn so that its end point follows the mouse cursor. If neither
        mode is selected, pan the image. event is a QGraphicsSceneMouseEvent
        object containing the coordinates of the mouse position.
        """

        # if the reference line is being drawn, redraw its line attribute so
        # that its end matches the current mouse location
        if self.angleRefLine.isBeingDrawn():
            self.angleRefLine.drawLine(
                event.pos().x(), event.pos().y(), self.lineWidth, self.scene)

        # if neither mode is currently selected, translate the image so that
        # the pixel under the mouse cursor follows the mouse (i.e. pan the
        # image)
        if not (self.placeMarkerButton.isChecked()
                or self.drawRefButton.isChecked()):
            # get the graphics view's scroll bar objects
            hbar = self.sceneView.horizontalScrollBar()
            vbar = self.sceneView.verticalScrollBar()
            # determine the difference between the current mouse event postion
            # and the previous mouse event position (in graphics view
            # coordinates)
            delta = self.sceneView.mapFromScene(
                        event.pos()) - self.sceneView.lastMousePos
            # store the position of this mouse event
            self.sceneView.lastMousePos = self.sceneView.mapFromScene(
                                              event.pos())
            # shift the scroll bars by the difference in mouse event positions
            hbar.setValue(hbar.value() - delta.x())
            vbar.setValue(vbar.value() - delta.y())

    ##############################
    # File Dialog Event Handlers
    ##############################
    def openImage(self, fileName=None):
        """If fileName, a string containing the complete location of an image
        is passed, open that image. Otherwise have the user select the image
        to open via file dialog.
        """

        # if no file name was given, open file dialog to obtain image file name
        if not fileName:
            fileName = QtWidgets.QFileDialog.getOpenFileName(
                None, "Open File", QtCore.QDir.currentPath(),
                "Images (*.png *.jpg);;All Files (*)")[0]

        # load the image into sceneImage
        if not fileName:
            return False # image not loaded successfully
        else:
            image = self.sceneImage.load(fileName)
        if not image:
            self.displayMessage(
                "NOTICE: Cannot open file as image: {}.".format(fileName))
            return False # image not loaded successfully

        # store the image file name
        self.imageFileName = fileName

        # resize the graphics scene to the loaded image dimensions
        self.scene.setSceneRect(
            0, 0, self.sceneImage.width(), self.sceneImage.height())

        # create a pixmap from the loaded image
        self.scenePixmap.setPixmap(QtGui.QPixmap.fromImage(self.sceneImage))

        # set keyboard focus to the graphics view
        self.sceneView.setFocus()

        # reset the application
        self.reset()

        # an image was successfully opened
        return True

    def saveSession(self):
        """Save analysis session to a .json file selected by the user via file
        dialog.
        """

        # if no image has been opened, return
        if self.scenePixmap.pixmap().isNull():
            self.displayMessage("NOTICE: Nothing to save.")
            return

        # open file dialog for selecting a file to save to
        fileName = QtWidgets.QFileDialog.getSaveFileName(
            None, "Save Session", "./untitled.json",
            "HEP Track Analysis (*.json);;All Files (*)")[0]

        # return if no file was selected
        if not fileName:
            return

        else:
            # initialize a dictionary for the data we want to save
            saveData = {}
            
            # store the image's file location
            saveData['imageFileName'] = self.imageFileName

            # if there are any markers in the marker list, store their data
            if self.markerList.count() > 0:
                # store the marker data in a list to preserve order
                points = []
                for row in range(self.markerList.count()):
                    point = self.markerList.item(row)
                    pointDict = {}
                    # save the marker designation and the coordinates
                    pointDict['designation'] = point.designation
                    pointDict['x'] = point.ellipse.rect().center().x()
                    pointDict['y'] = point.ellipse.rect().center().y()
                    points.append(pointDict)
                saveData["points"] = points

            # store the dL if it is not empty or 0
            if self.dlLineEdit.text() not in ["0", ""]:
                saveData['dl'] = self.dlLineEdit.text()

            # store the coordinates of the initial and final points of the
            # reference line
            if self.angleRefLine.finalPoint:
                initialPointDict = {}
                initialPointDict[
                    'x'] = self.angleRefLine.initialPoint.rect().center().x()
                initialPointDict[
                    'y'] = self.angleRefLine.initialPoint.rect().center().y()
                saveData['refInitialPoint'] = initialPointDict
                finalPointDict = {}
                finalPointDict[
                    'x'] = self.angleRefLine.finalPoint.rect().center().x()
                finalPointDict[
                    'y'] = self.angleRefLine.finalPoint.rect().center().y()
                saveData['refFinalPoint'] = finalPointDict

            # serialize the save data dictionary and save to file
            with open(fileName, 'w') as saveFile:
                json.dump(saveData, saveFile, indent=4)

    def loadSession(self):
        """Load an analysis session from a .json file selected by the user via
        file dialog.
        """

        # open file dialog for selecting a file to load from
        fileName = QtWidgets.QFileDialog.getOpenFileName(
            None, "Load Session", QtCore.QDir.currentPath(),
            "HEP Track Analysis (*.json);;All Files (*)")[0]

        # return if no file was selected
        if not fileName:
            return

        else:
            # open the file contents
            with open(fileName, 'r') as loadFile:
                # try to load the file contents as a JSON formatted object
                try:
                    loadData = json.load(loadFile)
                except:
                    self.displayMessage("NOTICE: Invalid JSON file: {}".format(
                                                                     fileName))
                    return

                # get the image filename from the saved session data
                imageFileName = loadData.get('imageFileName')

                # if the image filename is missing, return
                if not imageFileName:
                    self.displayMessage("NOTICE: No image file name found in saved session data: {}".format(fileName))
                    return

                # try to open the image. If it fails to open, return
                opened = self.openImage(imageFileName)
                if not opened:
                    return

                # get the track marker data from the saved session
                points = loadData.get('points')
                if points:
                    # add each track marker to the marker list
                    for point in points:
                        pointDesignation = point["designation"]
                        x = point['x']
                        y = point['y']
                        addedMarker = self.markerList.addMarker(
                                          x, y, self.pointSize,
                                          self.lineWidth, self.scene)
                        # set the appropriate designation for each marker
                        addedMarker.setDesignation(pointDesignation)

                # get the dl data from the saved session
                dl = loadData.get('dl')
                # if it is a float, set it to the dl text box value
                try:
                    float(dl)
                    self.dlLineEdit.setText(dl)
                except (ValueError, TypeError):
                    pass

                # get the data for the initial and final points for the
                # reference line
                refInitialPoint = loadData.get('refInitialPoint')
                refFinalPoint = loadData.get('refFinalPoint')
                if refInitialPoint and refFinalPoint:
                    # set the initial point, line and final point of the
                    # reference line
                    self.angleRefLine.setInitialPoint(
                        refInitialPoint['x'], refInitialPoint['y'],
                        self.pointSize, self.lineWidth, self.scene)
                    self.angleRefLine.drawLine(
                        refFinalPoint['x'], refFinalPoint['y'],
                        self.lineWidth, self.scene)
                    self.angleRefLine.setFinalPoint(
                        refFinalPoint['x'], refFinalPoint['y'],
                        self.pointSize, self.lineWidth, self.scene)

    def saveScreenshot(self):
        """Save the currently visible part of the graphics scene to an
        image.
        """

        # if no image has been opened, return
        if self.scenePixmap.pixmap().isNull():
            self.displayMessage(
                "NOTICE: There is nothing to take screenshot of.")
            return

        # grab the currently visible portion of the graphics scene and put it
        # in a pixmap
        screenshot = self.sceneView.grab()

        # open a file dialog for selecting a file to save to
        fileName = QtWidgets.QFileDialog.getSaveFileName(
            None, "Save Screenshot", "./untitled.png",
            "PNG (*.png);;JPEG (*.jpg);;TIFF (*.tiff *.tif)")[0]

        # return if no file was selected
        if not fileName:
            return

        # try to save the pixmap to the selected file
        if not screenshot.save(fileName):
            self.displayMessage("NOTICE: Unable to save screenshot.")

    ##############################
    # Zoom Events Handlers
    ##############################
    def zoomIn(self):
        """Scale the image by ZOOMINFACTOR."""
        self.scaleImage(constants.ZOOMINFACTOR)

    def zoomOut(self):
        """Scale the image by ZOOMOUTFACTOR."""
        self.scaleImage(constants.ZOOMOUTFACTOR)

    def scaleImage(self, factor):
        """Scale the graphics view and all graphics items drawn on it by
        factor.
        """
        # update the current zoom level
        self.zoomFactor = self.zoomFactor * factor
        
        # scale the graphics view
        self.sceneView.scale(factor, factor)

        # scale the point size and line width state variables
        self.pointSize /= factor
        self.lineWidth /= factor

        # scale all graphics items drawn on the graphics scene
        # using the updated point size and line width
        self.markerList.rescale(self.pointSize, self.lineWidth)
        self.momentumArc.rescale(self.lineWidth)
        self.angleRefLine.rescale(self.pointSize, self.lineWidth)
        if self.tangentLine:
            self.tangentLine.rescale(self.lineWidth)

    ###################################
    # Calculation Button Event Handlers
    ###################################
    def calcTrackMomentum(self):
        """ Perform the 'Calculate Track Momentum' behaviour.
        
        A circle is fit to the track markers in markerList. The calculated
        radius, momentum, length, and errors are then displayed.  The fitted
        momentum arc is drawn over the image.
        """
        # User input checks:
        if self.markerList.count() < 3:
            self.displayMessage("NOTICE: At least 3 points are required to calculate momentum.")
            return
        if not self.markerList.getStartPoint():
            self.displayMessage(
                "NOTICE: Track start point must be selected first.")
            return
        if not self.markerList.getEndPoint():
            self.displayMessage(
                "NOTICE: Track end point must be selected first.")
            return
        try:
            dl = float(self.dlLineEdit.text())
        except ValueError:
            dl = 0.

        # Do the calculations
        self.fittedCircle = circlefit.fitCircle(self.markerList)
        self.fittedCircle['cmRadius'] = self.fittedCircle['radius'] * constants.CMPERPX
        self.fittedCircle['cmRadiusErr'] = self.fittedCircle['radiusErr'] * constants.CMPERPX
        self.fittedCircle['cmRadiusCal'] = self.fittedCircle['radius'] * constants.ERRCMPERPX

        startAngle = self.markerList.getStartPoint().getAngle(
                         (self.fittedCircle['centerX'],
                          self.fittedCircle['centerY']))
        spanAngle = self.markerList.getEndPoint().getAngle(
                        (self.fittedCircle['centerX'], 
                         self.fittedCircle['centerY']), 
                        self.markerList.getStartPoint())
        trackLengthPx = self.fittedCircle['radius'] * spanAngle * (math.pi / 180)

        trackLengthCm = trackLengthPx * constants.CMPERPX
        trackLengthCmErr = trackLengthPx * constants.ERRCMPERPX

        momentum, mStatErr, mCalErr = circlefit.calc_momentum(
                self.fittedCircle['radius'],
                self.fittedCircle['radiusErr'])

        # draw the new momentum arc (removing previous drawn elements)
        if self.tangentLine:
            self.tangentLine.scene().removeItem(self.tangentLine)
            self.tangentLine = None
        self.momentumArc.draw(
            self.fittedCircle['centerX'], self.fittedCircle['centerY'],
            self.fittedCircle['radius'], startAngle, spanAngle, dl,
            self.lineWidth, self.scene)

        # Print the calculation results
        # raw circle parameters
        cParamMessage = """---Fitted Circle---
        Center (x coord):\t{centerX:.5f} +/- {centerXErr:.5f} [px]           
        Center (y coord):\t{centerY:.5f} +/- {centerYErr:.5f} [px]
        Radius (px):\t{radius:.5f} +/- {radiusErr:.5f} [px]
        Radius (cm):\t{cmRadius:.5f} +/- {cmRadiusErr:.5f} (Stat) +/- {cmRadiusCal:.5f} (Cal) [cm]
        """
        self.displayMessage(cParamMessage.format(**self.fittedCircle))

        mMessage = """---Track Momentum---
        Track Momentum:\t{:.5f} +/- {:.5f} (Stat) +/- {:.5f} (Cal) [MeV/c]
        """
        self.displayMessage(mMessage.format(momentum, mStatErr, mCalErr))

        trackLengthMessage = """---Track Length---
        Track Length (px):\t{:.5f} [px]
        Track Length (cm):\t{:.5f} +/- {:.5f} [cm]
        """
        self.displayMessage(trackLengthMessage.format(
            trackLengthPx, trackLengthCm, trackLengthCmErr))

    def calcOptDensity(self):
        """Calculate the optical density of a track and print it to the console.
        """
        # Check the user's inputs
        if not self.momentumArc.centralArc:
            self.displayMessage(
                "NOTICE: Track momentum must be calculated first.")
            return
        try:
            dl = float(self.dlLineEdit.text())
        except ValueError:
            dl = 0
        if dl == 0:
            self.displayMessage("NOTICE: dL must be non-zero.")
            return

        ## Calculations
        # compute the total blackness of all the pixels contained within the
        # portion of the sceneImage that is covered by the momentum arc
        # note: ArcItems have start and span angles in units of millionths of a
        # degree, so divide them by 1e6
        blackness, blacknessErr = optdensity.calcBlackness(
            self.sceneImage, self.fittedCircle, dl,
            self.momentumArc.centralArc.startAngle() / 1e6,
            self.momentumArc.centralArc.spanAngle() / 1e6)

        trackLengthPx = self.fittedCircle['radius'] * \
                 self.momentumArc.centralArc.spanAngle() / 1e6 * (math.pi / 180)

        trackLengthCm = trackLengthPx * constants.CMPERPX
        trackLengthCmErr = trackLengthPx * constants.ERRCMPERPX

        optDensity = blackness / trackLengthCm
        optDensityErr = optDensity * (
                (trackLengthCmErr / trackLengthCm)**2 + \
                (blacknessErr / blackness)**2)**0.5

        # print the optical density to the console
        message = """---Optical Density---
        Optical density:\t{:.5f} +/- {:.5f} [1/cm] (with dL={})
        """
        self.displayMessage(message.format(optDensity, optDensityErr, dl))

    def calcAngle(self):
        """Calculate the starting angle and print it to the console.

        The angle is calculated between the reference angle, and The tangent
        line at the designated start point of the circle.
        """
        # Check user inputs
        if not self.momentumArc.centralArc:
            self.displayMessage(
                "NOTICE: Track momentum must be calculated first.")
            return
        if not self.markerList.getStartPoint():
            self.displayMessage(
                "NOTICE: Start track point must be selected first.")
            return
        if not self.angleRefLine.finalPoint:
            self.displayMessage(
                "NOTICE: Angle Reference Line must be drawn first.")
            return

        # Calculations
        tangentLine, tangentErrA, tangentErrB = anglecalc.tangentCalc(
                          self.fittedCircle, self.markerList.getStartPoint())
        angle, angleErr = anglecalc.openingAngle(tangentLine,
                                                 tangentErrA, tangentErrB,
                                                 self.angleRefLine)

        # Drawing
        if self.tangentLine:
            self.tangentLine.scene().removeItem(self.tangentLine)
        self.tangentLine = tangent.TangentLine(tangentLine,
                                               self.lineWidth, self.scene)

        # print the opening angle to the console
        message = """---Opening Angle---
        Opening Angle:\t{:.5f} +/- {:.5f}
        """
        self.displayMessage(message.format(angle, angleErr))

    ##############################
    # Mode Change Event Handlers
    ##############################
    def placeMarkerButtonFunc(self):
        """Ensure that the angle reference drawing mode is not selected at the
        same time as the track marker placement mode.
        """

        self.drawRefButton.setChecked(False)

    def drawRefButtonFunc(self):
        """Ensure that the track marker placement mode is not selected at the
        same time as the angle reference drawing mode.
        """

        self.placeMarkerButton.setChecked(False)

    ##############################
    # Other Event Handlers
    ##############################
    def dLEdited(self, newDL):
        """Update the outer and inner arcs of the momentum arc to reflect
        newDL, the new value in the dL text box.
        """

        # if the dL text box is empty (and so newDL is an empty string),
        # return
        if not newDL:
            return

        # otherwise convert newDL to a float and update the arcs
        dl = float(newDL)
        self.momentumArc.updateArcs(dl)

    def highlightPoint(self):
        """Highlight the track marker that is currently selected."""

        self.markerList.highlightCurrent()

    ##############################
    # Reset Event Handler
    ##############################
    def reset(self):
        """Remove all graphics items drawn on the graphics scene, clear all
        messages from the console, reset image zoom and reset dL.
        """

        # remove all points, arcs and lines from the graphics scene
        self.markerList.empty()
        self.angleRefLine.reset()
        self.momentumArc.reset()
        if self.tangentLine:
            self.tangentLine.scene().removeItem(self.tangentLine)
            self.tangentLine = None

        # clear the console
        self.consoleTextBrowser.clear()

        # reset the dL to 0
        self.dlLineEdit.setText("0")

        # reset any image zoom
        self.scaleImage(1 / self.zoomFactor)

        # scale the image so that it fills as much of the graphics view as it
        # can without requiring scroll bars
        if not self.sceneImage.isNull():
            # determine how many times smaller (or larger) the graphics view
            # height is than the image height. Same for the widths.
            # note: the graphics view object has 2 extra pixels of height and
            # width that the image does not occupy
            heightRatio = (self.sceneView.height()-2) / \
                           self.sceneImage.height()
            widthRatio = (self.sceneView.width()-2) / \
                          self.sceneImage.width()

            if heightRatio < widthRatio:
                scaleFactor = heightRatio
            else:
                scaleFactor = widthRatio

            self.scaleImage(scaleFactor)

    ##############################
    # Helper Methods
    ##############################
    def displayMessage(self, msg):
        """Write msg, a string, along with the message number to the console.
        """
        self.msgNumber += 1
        msg = "[{}]  {}".format(self.msgNumber, msg)
        self.consoleTextBrowser.append(msg)



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
from PyQt5 import QtCore, QtGui, QtWidgets
from traxis.graphics import markers, angleref, fittedarc


class GuiSkeleton(QtWidgets.QWidget):

    """The topmost widget which places all the GUI's widgets onto itself upon
    initialization.
    """

    def __init__(self):
        """Setup the base user interface - create layouts and place widgets
        and labels.
        """

        super().__init__()

        # main layout of the skeleton
        self.mainLayout = QtWidgets.QVBoxLayout(self)

        # layout for the top portion of the user interface
        self.topWidget = QtWidgets.QWidget()
        self.mainLayout.addWidget(self.topWidget)
        # the top portion should have a fixed height, just big enough to fit
        # all of its contents
        self.topWidget.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.topUiLayout = QtWidgets.QHBoxLayout(self.topWidget)
        # don't add any extra padding around the edges of this layout's widgets
        self.topUiLayout.setContentsMargins(0, 0, 0, 0)

        # track marker list GUI segment
        self.markerListLayout = QtWidgets.QVBoxLayout()  # marker list layout
        self.topUiLayout.addLayout(self.markerListLayout)

        self.markerListLabel = QtWidgets.QLabel(self)  # marker list label
        self.markerListLayout.addWidget(self.markerListLabel)
        self.markerListLabel.setText("Track Markers")

        self.markerList = markers.MarkerList(self)  # marker list widget
        self.markerListLayout.addWidget(self.markerList)
        # don't focus on this widget when clicked
        self.markerList.setFocusPolicy(QtCore.Qt.NoFocus)
        self.markerList.setFixedWidth(100)

        # first vertical GUI segment divider in top portion layout
        self.vLineDiv1 = QtWidgets.QFrame(self)  # vertical divider widget
        self.topUiLayout.addWidget(self.vLineDiv1)
        self.vLineDiv1.setFrameShape(QtWidgets.QFrame.VLine)
        self.vLineDiv1.setFrameShadow(QtWidgets.QFrame.Sunken)

        # "technical button" GUI segment
        # technical button segment layout
        self.techButtonLayout = QtWidgets.QVBoxLayout()
        self.topUiLayout.addLayout(self.techButtonLayout)

        self.resetButtonLabel = QtWidgets.QLabel(self)  # reset button label
        self.techButtonLayout.addWidget(self.resetButtonLabel)
        self.resetButtonLabel.setText("Reset Analysis")

        self.resetButton = QtWidgets.QPushButton(self)  # reset button widget
        self.techButtonLayout.addWidget(self.resetButton)
        # don't focus on this widget when clicked
        self.resetButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.resetButton.setText("Reset")
        self.resetButton.setToolTip(
            "Reset all the selected points and calculated variables")
        self.resetButton.setShortcut(QtGui.QKeySequence("R"))

        self.zoomLabel = QtWidgets.QLabel(self)  # zoom label
        self.techButtonLayout.addWidget(self.zoomLabel)
        self.zoomLabel.setText("Zoom")

        # horizontal layout for zoom buttons
        self.zoomLayout = QtWidgets.QHBoxLayout()
        self.techButtonLayout.addLayout(self.zoomLayout)

        # zoom in button widget
        self.zoomInButton = QtWidgets.QPushButton(self)
        self.zoomLayout.addWidget(self.zoomInButton)
        # don't focus on this widget when clicked
        self.zoomInButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.zoomInButton.setText("Zoom In")
        self.zoomInButton.setToolTip("Zoom into the picture")
        self.zoomInButton.setShortcut(QtGui.QKeySequence("Z"))

        # zoom out button widget
        self.zoomOutButton = QtWidgets.QPushButton(self)
        self.zoomLayout.addWidget(self.zoomOutButton)
        self.zoomOutButton.setFocusPolicy(QtCore.Qt.NoFocus)
        # don't focus on this widget when clicked
        self.zoomOutButton.setText("Zoom Out")
        self.zoomOutButton.setToolTip("Zoom out from the picture")
        self.zoomOutButton.setShortcut(QtGui.QKeySequence("X"))

        self.calcLabel = QtWidgets.QLabel(self)  # calculate label
        self.techButtonLayout.addWidget(self.calcLabel)
        self.calcLabel.setText("Calculate")

        # calculate momentum button widget
        self.calcMomentumButton = QtWidgets.QPushButton(self)
        self.techButtonLayout.addWidget(self.calcMomentumButton)
        # don't focus on this widget when clicked
        self.calcMomentumButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.calcMomentumButton.setText("Calculate Track Momentum")
        self.calcMomentumButton.setToolTip("Calculate Track momentum")
        self.calcMomentumButton.setShortcut(QtGui.QKeySequence("M"))

        # calculate optical density button widget
        self.calcDensityButton = QtWidgets.QPushButton(self)
        self.techButtonLayout.addWidget(self.calcDensityButton)
        # don't focus on this widget when clicked
        self.calcDensityButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.calcDensityButton.setText("Calculate Optical Density")
        self.calcDensityButton.setToolTip("Calculate Optical Density")
        self.calcDensityButton.setShortcut(QtGui.QKeySequence("N"))

        # calculate angle button widget
        self.calcAngleButton = QtWidgets.QPushButton(self)
        self.techButtonLayout.addWidget(self.calcAngleButton)
        # don't focus on this widget when clicked
        self.calcAngleButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.calcAngleButton.setText("Calculate Angle")
        self.calcAngleButton.setToolTip("Calculate Opening Angle")
        self.calcAngleButton.setShortcut(QtGui.QKeySequence("B"))

        # add stretch to segment to keep widgets together
        self.techButtonLayout.addStretch(0)

        # second vertical GUI segment divider in top portion layout
        self.vLineDiv2 = QtWidgets.QFrame(self)  # vertical divider widget
        self.topUiLayout.addWidget(self.vLineDiv2)
        self.vLineDiv2.setFrameShape(QtWidgets.QFrame.VLine)
        self.vLineDiv2.setFrameShadow(QtWidgets.QFrame.Sunken)

        # "user selection" GUI segment
        # user seletion segment layout
        self.userSelectionLayout = QtWidgets.QVBoxLayout()
        self.topUiLayout.addLayout(self.userSelectionLayout)

        self.openSaveLabel = QtWidgets.QLabel(self)  # open/save label
        self.userSelectionLayout.addWidget(self.openSaveLabel)
        self.openSaveLabel.setText("Open/Save")

        self.openImageButton = QtWidgets.QPushButton(
            self)  # open image button widget
        self.userSelectionLayout.addWidget(self.openImageButton)
        # don't focus on this widget when clicked
        self.openImageButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.openImageButton.setText("Open Image")
        self.openImageButton.setToolTip("Open image for analysis")
        self.openImageButton.setShortcut(QtGui.QKeySequence("O"))

        # horizontal layout for save and load buttons
        self.saveLayout = QtWidgets.QHBoxLayout()
        self.userSelectionLayout.addLayout(self.saveLayout)

        # save session button widget
        self.saveSessionButton = QtWidgets.QPushButton(self)
        self.saveLayout.addWidget(self.saveSessionButton)
        # don't focus on this widget when clicked
        self.saveSessionButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.saveSessionButton.setText("Save")
        self.saveSessionButton.setToolTip("Save current analysis session")
        
        # load session button widget
        self.loadSessionButton = QtWidgets.QPushButton(self)
        self.saveLayout.addWidget(self.loadSessionButton)
        # don't focus on this widget when clicked
        self.loadSessionButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.loadSessionButton.setText("Load")
        self.loadSessionButton.setToolTip(
            "Load a previously saved analysis session")

        # screenshot button widget
        self.screenshotButton = QtWidgets.QPushButton(self)
        self.userSelectionLayout.addWidget(self.screenshotButton)
        # don't focus on this widget when clicked
        self.screenshotButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.screenshotButton.setText("Save Screenshot")
        self.screenshotButton.setToolTip(
            "Take a screenshot of the scroll area contents and save to image")

        self.modeLabel = QtWidgets.QLabel(self) # mode label
        self.userSelectionLayout.addWidget(self.modeLabel)
        self.modeLabel.setText("Mode")

        # place marker mode button widget
        self.placeMarkerButton = QtWidgets.QPushButton(self)
        self.userSelectionLayout.addWidget(self.placeMarkerButton)
        # make the button checkable (i.e. stays depressed when clicked)
        self.placeMarkerButton.setCheckable(True)
        # don't focus on this widget when clicked
        self.placeMarkerButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.placeMarkerButton.setText("[Mode] Place Track Markers")
        self.placeMarkerButton.setToolTip(
            "Enter mode for placing markers on loaded image.")
        self.placeMarkerButton.setShortcut("P")

        # draw angle reference mode button widget
        self.drawRefButton = QtWidgets.QPushButton(self)
        self.userSelectionLayout.addWidget(self.drawRefButton)
        # make the button checkable (i.e. stays depressed when clicked)
        self.drawRefButton.setCheckable(True)
        # don't focus on this widget when clicked
        self.drawRefButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.drawRefButton.setText("[Mode] Draw Angle Reference")
        self.drawRefButton.setToolTip(
            "Enter mode for drawing angle reference on loaded image.")
        self.drawRefButton.setShortcut("L")

        # dl form layout
        self.dlFormLayout = QtWidgets.QFormLayout()
        self.userSelectionLayout.addLayout(self.dlFormLayout)

        # dl label
        self.dlLabel = QtWidgets.QLabel(self)
        self.dlFormLayout.setWidget(
            0, QtWidgets.QFormLayout.LabelRole, self.dlLabel)
        self.dlLabel.setText("Set dL")

        # dl text box (line edit) widget
        self.dlLineEdit = QtWidgets.QLineEdit(self)
        # fix the size of the text box
        self.dlLineEdit.setSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.dlFormLayout.setWidget(
            0, QtWidgets.QFormLayout.FieldRole, self.dlLineEdit)
        # set the dL value to 0 by default
        self.dlLineEdit.setText("0")
        # validate the contents of the text box so that only floats can
        # be entered
        self.dlLineEdit.setValidator(
            QtGui.QRegExpValidator(QtCore.QRegExp('[0-9]+\.?[0-9]*')))

        # add stretch to segment to keep widgets together
        self.userSelectionLayout.addStretch(0)

        # third vertical GUI segment divider in top portion layout
        self.vLineDiv3 = QtWidgets.QFrame(self)  # vertical divider widget
        self.topUiLayout.addWidget(self.vLineDiv3)
        self.vLineDiv3.setFrameShape(QtWidgets.QFrame.VLine)
        self.vLineDiv3.setFrameShadow(QtWidgets.QFrame.Sunken)

        # console GUI segment
        self.consoleLayout = QtWidgets.QVBoxLayout()  # console segment layout
        self.topUiLayout.addLayout(self.consoleLayout)

        self.consoleLabel = QtWidgets.QLabel(self)  # console label
        self.consoleLayout.addWidget(self.consoleLabel)
        self.consoleLabel.setText("Console")

        # console text browser widget
        self.consoleTextBrowser = QtWidgets.QTextBrowser(self)
        self.consoleLayout.addWidget(self.consoleTextBrowser)
        self.consoleTextBrowser.setMinimumWidth(100)

        # determine the height of the "user selection" GUI segment (the
        # tallest segment in the top portion of the GUI)
        userSelectionHeight = \
            self.userSelectionLayout.minimumSize().height() - \
            self.openSaveLabel.height()
        # fix the height of the marker list and console to the "user selection"
        # GUI segment height
        self.markerList.setFixedHeight(userSelectionHeight)
        self.consoleTextBrowser.setFixedHeight(userSelectionHeight)

        # horizontal GUI segment divider between top portion and bottom
        # portion of the GUI
        self.hLineDiv = QtWidgets.QFrame(self)
        self.mainLayout.addWidget(self.hLineDiv)
        self.hLineDiv.setFrameShape(QtWidgets.QFrame.HLine)
        self.hLineDiv.setFrameShadow(QtWidgets.QFrame.Sunken)

        # layout for the bottom portion of the user interface
        self.bottomUiLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.addLayout(self.bottomUiLayout)

        # create a graphics scene on which images and all graphics will be
        # displayed
        self.scene = QtWidgets.QGraphicsScene()
        # the graphics view is the widget that actually displays the contents
        # of the graphics scene
        self.sceneView = QtWidgets.QGraphicsView(self.scene, self)
        self.bottomUiLayout.addWidget(self.sceneView)
        # set a minimum size for the scene view
        self.sceneView.setMinimumWidth(900)
        self.sceneView.setMinimumHeight(400)

        # instantiate QImage and PixmapItem
        self.sceneImage = QtGui.QImage()
        self.scenePixmap = QtWidgets.QGraphicsPixmapItem()
        self.scene.addItem(self.scenePixmap)

        # instantiate reference line and momentum arc objects
        self.angleRefLine = angleref.ReferenceLine()
        self.momentumArc = fittedarc.MomentumArc()
 
        # set the tangentLine attribute to None initially so that there is
        # something to check for when a tangent has not been drawn yet
        self.tangentLine = None



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from PyQt5 import QtGui


def calcBlackness(image, circleParams, dL, startAngle, spanAngle):
    """Given image, a QImage, a circle defined by circleParams (a dict
    containing the radius and centre coordinates of a circle), a
    startAngle and spanAngle (both in degrees) specifying an arc of that
    circle, and dL, the 'infinitesimal' thickness of a polar rectangle
    surrounding that arc, return the sum of the blackness of all the pixels
    in image that are contained within the polar rectangle along with an
    error on the blackness.
    """

    # assume 1 px error on the dL
    dLErr = 1

    # initialize a set of distinct points that are contained within the polar
    # rectangle surrounding an arc of the circle defined by circleParams
    # from startAngle to startAngle + spanAngle and with radial thickness
    # 2*dL+1
    pointSet = set()
    # initialize also a set of distinct points that are in the polar rectangles
    # of thickness dLErr that lie radially just above and just below the
    # previously defined polar rectangle
    errPointSet = set()

    # loop over all the points in the combined area of the three polar
    # rectangles, adding each point to the appropriate set
    radii = np.linspace(circleParams['radius'] - dL - dLErr,
                        circleParams['radius'] + dL + dLErr,
                        int(2 * (dL + dLErr) + 1))
    for r in radii:
        # for the number of angles to generate, use twice the length of the
        # arc in pixels to ensure every pixel in the region is covered
        angles = np.linspace(startAngle, startAngle + spanAngle,
                             int(2 * r * spanAngle * (np.pi / 180)))
        # compute cos and sin for each angle
        cosAngles = np.cos(angles * (np.pi / 180))
        sinAngles = np.sin(angles * (np.pi / 180))
        for angleIndex in range(len(angles)):
            # get the x and y coordinates of the point
            x = circleParams['centerX'] + r * cosAngles[angleIndex]
            # note: y values increase going down
            y = circleParams['centerY'] - r * sinAngles[angleIndex]
            # add the point to the appropriate set
            if r < (circleParams['radius'] - dL) or r > (circleParams['radius'] + dL):
                errPointSet.add((int(x), int(y)))
            else:
                pointSet.add((int(x), int(y)))

    # sum up the black colour components of pixels located at the points in
    # pointSet
    blackness = 0
    for point in pointSet:
        pixel = image.pixel(point[0], point[1])
        blackness += QtGui.QColor(pixel).blackF()

    # sum up the black colour components of pixels located at the points in
    # errPointSet
    errBlackness = 0
    for point in errPointSet:
        pixel = image.pixel(point[0], point[1])
        errBlackness += QtGui.QColor(pixel).blackF()

    return blackness, errBlackness



from setuptools import setup, find_packages

import traxis

setup(
    name='traxis',
    version=traxis.__version__,
    url='https://gitlab.physics.utoronto.ca/advanced-lab/traxis',
    author='Syed Haider Abidi, Nooruddin Ahmed, Christopher Dydul',
    maintainer='John Ladan',
    maintainer_email='jladan@physics.utoronto.ca',
    # Package info
    packages=find_packages(),
    install_requires=['numpy', 'scipy', 'PyQt5'],
    package_data={
        "": ["README.md", '*.png']
    },
    license='GPL3',
    description='An application to measure bubble-chamber images for the UofT advanced physics lab.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    scripts=['runtraxis'],
    entry_points={
        "gui_scripts": [
            "traxis = traxis.__main__:main"
        ]
    }, 
)



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtWidgets, QtGui, QtCore
from traxis import constants


class TangentLine(QtWidgets.QGraphicsLineItem):
    """Tangent line class. This class subclasses QGraphicsLineItem, allowing
    an object to be instantiated with a pen width and a graphics scene to
    which it will be added in addition to a QLineF object. It also adds a
    method to facilitate rescaling of the tangent line.
    """

    def __init__(self, qlinef, width, scene):
        """Instantiate a TangentLine object with qlinef, a QLineF object,
        width, the width of the pen of the tangent line, and scene, a
        QGraphicsScene to which the tangent line will be added.
        """

        # call the __init__ method of the QGraphicsLineItem superclass, passing
        # the QLineF object
        super().__init__(qlinef)

        # set a minimum pen width
        if width < 1:
            width = 1
        width = int(width)

        # create a pen for the tangent line using the tangent colour
        linePen = QtGui.QPen(constants.TANGENTLINECOLOR)
        # set the width of the pen to width
        linePen.setWidth(width)
        # set the newly created pen as the tangent line's pen
        self.setPen(linePen)

        # add the tangent line to the graphics scene
        scene.addItem(self)

    def rescale(self, width):
        """Set the pen width of the tangent line to width."""

        # set a minimum pen width
        if width < 1:
            width = 1
        width = int(width)

        # create a new pen, starting from the existing pen of the tangent line
        newLinePen = self.pen()
        # set the width of the new pen
        newLinePen.setWidth(width)
        # set the resized pen as the tangent line's pen
        self.setPen(newLinePen)



# Copyright (C) 2014 Syed Haider Abidi, Nooruddin Ahmed and Christopher Dydula
#
# This file is part of traxis.
#
# traxis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# traxis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with traxis.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtGui, QtCore, QtWidgets
from traxis import constants


class ArcItem(QtWidgets.QGraphicsEllipseItem):
    """This class is a QGraphicsEllipseItem with a custom paint method.
    Instead a drawing an ellipse or a pie, an ArcItem object draws an arc
    when added to a graphics scene. The start and span angles of the ArcItem
    are specified in millionths of a degree instead of 16ths of a degree
    for greater precision when drawing. The start and span angles of an ArcItem
    object must be specified before it is added to a graphics scene.
    """

    def paint(self, painter, option, widget=0):
        """Reimplement the paint method so that it draws arcs. painter is a
        QPainter object. option and widget are included in the parameter
        list because a graphics item paint method expects them but they are
        not used here."""

        # set the pen of the painter to the ArcItem object's pen
        painter.setPen(self.pen())
        # set the brush of the painter to the ArcItem object's brush
        painter.setBrush(self.brush())
        # draw an arc, using the ArcItem object's rect, startAngle and
        # spanAngle
        path = QtGui.QPainterPath()
        path.arcMoveTo(self.rect().normalized(), self.startAngle() / 1e6)
        path.arcTo(self.rect().normalized(),
                   self.startAngle() / 1e6, self.spanAngle() / 1e6)
        painter.drawPath(path)

    def shape(self):
        """Reimplement the shape method to reflect the fact that ArcItem
        object start and span angles are in millionths of a degree instead
        of 16ths of a degree. This method is exactly the same of the default
        shape() method but with all instances of '16' replaced with '1e6'.
        """

        path = QtGui.QPainterPath()
        if self.rect().isNull():
            return path

        if self.spanAngle != 360 * 1e6:
            path.moveTo(self.rect().center())
            path.arcTo(self.rect(),
                       self.startAngle() / 1e6, self.spanAngle() / 1e6)
        else:
             path.addEllipse(self.rect())

        if path == QtGui.QPainterPath():
            return path
        ps = QtGui.QPainterPathStroker()
        ps.setCapStyle(self.pen().capStyle())
        if self.pen().widthF() <= 0.0:
            ps.setWidth(0.00000001)
        else:
            ps.setWidth(self.pen().widthF())
        ps.setJoinStyle(self.pen().joinStyle())
        ps.setMiterLimit(self.pen().miterLimit())
        p = ps.createStroke(path)
        p.addPath(path)
        return p

class MomentumArc(object):
    """Momentum arc class. This class is a container for three ArcItem objects
    which together form the momentum arc. This class implements methods for
    manipulating the ArcItems it contains such that they function as a unit.
    A MomentumArc object is to be added to a GUI as one would add a QWidget.
    """

    def __init__(self):
        """Initialize this object's attributes, setting them to None."""

        # the attributes are all ArcItems, each with the same center, or all
        # None, if the momentum arc has not yet been drawn. The radius of the
        # outer arc is to be greater than or equal to the central arc. The
        # radius of the inner arc is to be less than or equal to the central
        # arc.
        self.centralArc = None
        self.outerArc = None
        self.innerArc = None

    def draw(self, centerX, centerY, radius, startAngle,
             spanAngle, dl, width, scene):
        """Draw the momentum arc and bounding arcs.

        First, reset the momentum arc if it has already been drawn. Then 
        create three ArcItems, one with radius radius, the second with radius
        radius+dl and the third with radius radius-dl. Set each arc's center to
        (centerX, centerY), the start angle to startAngle and the span
        angle to spanAngle, both given in degrees. Set each arc's pen width to
        width and add each arc to scene, a QGraphicsScene.

        Parameters
        ----------
        centerX : int
        centerY : int
            center of the fitted circle
        radius : float
            radius of the fitted circle
        startAngle : float (degrees)
            starting angle of the fitted arc
        spanAngle : float (degrees)
            length of the arc in degrees
        dl : int
            width of the region of interest for optical density
        scene : QGraphicsScene
            scene on which to draw the arcs
        """

        # reset the momentum arc object
        self.reset()

        # set a minimum pen width
        if width < 1:
            width = 1
        width = int(width)

        # create a pen for the arcs using the momentum arc colour
        momentumPen = QtGui.QPen(constants.ARCCOLOR)
        # set the width of the pen to width
        momentumPen.setWidth(width)

        # create a rect for the central arc with the given center
        # coordinates and radius (for an arc with radius r, the rect will
        # have width and height 2*r)
        centralRect = QtCore.QRectF(centerX, centerY, 2 * radius, 2 * radius)
        # when instantiating a rect, the given coordinates are used for the
        # top-left corner so we must explicitly move the rect's center to the
        # desired coordinates
        centralRect.moveCenter(QtCore.QPointF(centerX, centerY))
        # create an ArcItem using the rect created above. Set it
        # as the momentum arc's centralArc attribute.
        self.centralArc = ArcItem(centralRect)
        # set the start and span angles of the central arc to the given values.
        # the arc item's angle unit is millionths of a degree so multiply the
        # given angles (which are in degrees) by 1e6
        self.centralArc.setStartAngle(int(1e6 * startAngle))
        self.centralArc.setSpanAngle(int(1e6 * spanAngle))
        # set the central arc's pen to the pen created above
        self.centralArc.setPen(momentumPen)

        # for the inner and outer arcs, use a dash-dot-line pen style
        momentumPen.setStyle(QtCore.Qt.DashDotLine)

        # create a rect for the outer arc with the given center
        # coordinates. Set the radius to radius+dl (for an arc with radius
        # r, the rect will have width and height 2*r)
        outerRect = QtCore.QRectF(
            centerX, centerY, 2 * (radius + dl), 2 * (radius + dl))
        # when instantiating a rect, the given coordinates are used for the
        # top-left corner so we must explicitly move the rect's center to the
        # desired coordinates
        outerRect.moveCenter(QtCore.QPointF(centerX, centerY))
        # create an ArcItem using the rect created above. Set it
        # as the momentum arc's outerArc attribute.
        self.outerArc = ArcItem(outerRect)
        # set the start and span angles of the outer arc to the given values.
        # the arc item's angle unit is millionths of a degree so multiply the
        # given angles (which are in degrees) by 1e6
        self.outerArc.setStartAngle(int(1e6 * startAngle))
        self.outerArc.setSpanAngle(int(1e6 * spanAngle))
        # set the outer arc's pen to the pen created above
        self.outerArc.setPen(momentumPen)

        # create a rect for the inner arc with the given center
        # coordinates. Set the radius to radius-dl (for an arc with radius
        # r, the rect will have width and height 2*r)
        innerRect = QtCore.QRectF(
            centerX, centerY, 2 * (radius - dl), 2 * (radius - dl))
        # when instantiating a rect, the given coordinates are used for the
        # top-left corner so we must explicitly move the rect's center to the
        # desired coordinates
        innerRect.moveCenter(QtCore.QPointF(centerX, centerY))
        # create an ArcItem using the rect created above. Set it
        # as the momentum arc's innerArc attribute.
        self.innerArc = ArcItem(innerRect)
        # set the start and span angles of the inner arc to the given values.
        # the arc item's angle unit is millionths of a degree so multiply the
        # given angles (which are in degrees) by 1e6
        self.innerArc.setStartAngle(int(1e6 * startAngle))
        self.innerArc.setSpanAngle(int(1e6 * spanAngle))
        # set the outer arc's pen to the pen created above
        self.innerArc.setPen(momentumPen)

        # add the three arcs to the graphics scene
        scene.addItem(self.centralArc)
        scene.addItem(self.outerArc)
        scene.addItem(self.innerArc)

    def updateArcs(self, dl):
        """Redraw the inner and outer arcs so that their radii are equal to
        central arc's +/- the new dl.
        """

        # if the arcs haven't been drawn yet, return
        if not self.centralArc:
            return

        # get the width of the central arc's rect (which is equal to twice its
        # radius)
        centralWidth = self.centralArc.rect().width()

        # create a new rect, starting from the existing rect of the outer arc
        newOuterRect = self.outerArc.rect()
        # set the width and height of the new rect, incorporating the new dl
        # (the width and height of the rect are twice the radius of the arc)
        newOuterRect.setWidth(centralWidth + 2 * dl)
        newOuterRect.setHeight(centralWidth + 2 * dl)
        # move the rect's center to match that of the original rect's (the
        # center gets shifted when resizing)
        newOuterRect.moveCenter(self.outerArc.rect().center())
        # set the resized rect as the outer arc's rect
        self.outerArc.setRect(newOuterRect)

        # create a new rect, starting from the existing rect of the inner arc
        newInnerRect = self.innerArc.rect()
        # set the width and height of the new rect, incorporating the new dl
        # (the width and height of the rect are twice the radius of the arc)
        newInnerRect.setWidth(centralWidth - 2 * dl)
        newInnerRect.setHeight(centralWidth - 2 * dl)
        # move the rect's center to match that of the original rect's (the
        # center gets shifted when resizing)
        newInnerRect.moveCenter(self.innerArc.rect().center())
        # set the resized rect as the inner arc's rect
        self.innerArc.setRect(newInnerRect)

    def rescale(self, width):
        """Set the pen width of the three arcs to width."""

        # set a minimum pen width
        if width < 1:
            width = 1
        width = int(width)

        # if the arcs haven't been drawn yet, return
        if not self.centralArc:
            return

        # create a new pen, starting from the existing pen of the central arc
        newPen = self.centralArc.pen()
        # set the width of the new pen
        newPen.setWidth(width)
        # set the resized pen as the central arc's pen
        self.centralArc.setPen(newPen)

        # for the inner and outer arcs, use a dash-dot-line pen style
        newPen.setStyle(QtCore.Qt.DashDotLine)
        # set the resized pen as the inner and outer arcs' pen
        self.outerArc.setPen(newPen)
        self.innerArc.setPen(newPen)

    def reset(self):
        """Remove the momentum arc's attributes from their graphics scene and
        reset them to None.
        """

        # if the central, inner and outer arcs exist, remove them from
        # their graphics scene
        if self.centralArc:
            self.centralArc.scene().removeItem(self.centralArc)
        if self.outerArc:
            self.outerArc.scene().removeItem(self.outerArc)
        if self.innerArc:
            self.innerArc.scene().removeItem(self.innerArc)

        # reset the momentum arc's attributes to None
        self.centralArc = None
        self.outerArc = None
        self.innerArc = None
